"""
Provide a module that describes permissions in the API.
"""

from enum import Enum
from typing import Union

from fastapi import HTTPException, status
from loguru import logger
from pydantic import EmailStr


class Permissions(str, Enum):
    """
    Describe the permissions that may be used for protecting Jobbergate routes.
    """

    APPLICATIONS_VIEW = "jobbergate:applications:view"
    APPLICATIONS_EDIT = "jobbergate:applications:edit"
    JOB_SCRIPTS_VIEW = "jobbergate:job-scripts:view"
    JOB_SCRIPTS_EDIT = "jobbergate:job-scripts:edit"
    JOB_SUBMISSIONS_VIEW = "jobbergate:job-submissions:view"
    JOB_SUBMISSIONS_EDIT = "jobbergate:job-submissions:edit"


def check_owner(
    owner_email: Union[str, EmailStr, None],
    requester_email: Union[str, EmailStr, None],
    entity_id: int,
    entity_name: str,
):
    """
    Assert ownership of an entity and raise a 403 exception with message on failure.
    """
    if requester_email != owner_email:
        message = (
            f"User {requester_email} does not own {entity_name} with id={entity_id}. "
            f"Only the {entity_name} owner ({owner_email}) "
            f"can modify this {entity_name}."
        )
        logger.error(message)
        raise HTTPException(status_code=status.HTTP_403_FORBIDDEN, detail=message)



"""
Database model for the JobScript resource.
"""
from sqlalchemy import Boolean, DateTime, ForeignKey, Integer, String, Table
from sqlalchemy.sql import False_, func
from sqlalchemy.sql.schema import Column

from jobbergate_api.metadata import metadata

job_scripts_table = Table(
    "job_scripts",
    metadata,
    Column("id", Integer, primary_key=True),
    Column("job_script_name", String, nullable=False, index=True),
    Column("job_script_description", String, default=""),
    Column("job_script_owner_email", String, nullable=False, index=True),
    Column("application_id", ForeignKey("applications.id"), nullable=True),
    Column("created_at", DateTime, nullable=False, default=func.now()),
    Column("updated_at", DateTime, nullable=False, default=func.now(), onupdate=func.now()),
    Column("is_archived", Boolean, nullable=False, default=False_()),
)

searchable_fields = [
    job_scripts_table.c.job_script_name,
    job_scripts_table.c.job_script_description,
    job_scripts_table.c.job_script_owner_email,
]

sortable_fields = [
    job_scripts_table.c.id,
    job_scripts_table.c.job_script_name,
    job_scripts_table.c.job_script_owner_email,
    job_scripts_table.c.application_id,
    job_scripts_table.c.created_at,
    job_scripts_table.c.updated_at,
]



"""
Router for the Application resource.
"""
from pathlib import PurePath
from typing import List, Optional, Union

from armasec import TokenPayload
from fastapi import APIRouter, Depends, File, Header, HTTPException, Query
from fastapi import Response as FastAPIResponse
from fastapi import UploadFile, status
from loguru import logger
from sqlalchemy import not_

from jobbergate_api.apps.applications.application_files import ApplicationFiles
from jobbergate_api.apps.applications.models import applications_table, searchable_fields, sortable_fields
from jobbergate_api.apps.applications.schemas import (
    ApplicationCreateRequest,
    ApplicationPartialResponse,
    ApplicationResponse,
    ApplicationUpdateRequest,
)
from jobbergate_api.apps.job_scripts.models import job_scripts_table
from jobbergate_api.apps.permissions import Permissions, check_owner
from jobbergate_api.config import settings
from jobbergate_api.pagination import Pagination, ok_response, package_response
from jobbergate_api.security import IdentityClaims, guard
from jobbergate_api.storage import (
    INTEGRITY_CHECK_EXCEPTIONS,
    database,
    render_sql,
    search_clause,
    sort_clause,
)

router = APIRouter()


async def _fetch_application_by_id(application_id: int) -> ApplicationPartialResponse:
    """
    Fetch an application from the database by its id.
    """
    select_query = applications_table.select().where(applications_table.c.id == application_id)
    raw_application = await database.fetch_one(select_query)

    if not raw_application:
        message = f"Application {application_id=} not found."
        logger.error(message)
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=message)

    return ApplicationPartialResponse.parse_obj(raw_application)


@router.post(
    "/applications",
    status_code=status.HTTP_201_CREATED,
    response_model=ApplicationPartialResponse,
    description="Endpoint for application creation",
)
async def applications_create(
    application: ApplicationCreateRequest,
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.APPLICATIONS_EDIT)),
):
    """
    Create new applications using an authenticated user token.
    """
    logger.debug(f"Creating {application=}")

    identity_claims = IdentityClaims.from_token_payload(token_payload)
    create_dict = dict(
        **application.dict(exclude_unset=True),
        application_owner_email=identity_claims.email,
    )

    logger.debug("Inserting application")

    try:
        insert_query = applications_table.insert().returning(applications_table)
        logger.trace(f"insert_query = {render_sql(insert_query)}")
        application_data = await database.fetch_one(query=insert_query, values=create_dict)

    except INTEGRITY_CHECK_EXCEPTIONS as e:
        logger.error(f"INTEGRITY_CHECK_EXCEPTIONS: {str(e)}")
        raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail=str(e))

    logger.debug(f"Applications created: {application_data=}")

    return application_data


@router.post(
    "/applications/{application_id}/upload",
    status_code=status.HTTP_201_CREATED,
    description="Endpoint for uploading application files.",
)
async def applications_upload(
    application_id: int = Query(..., description="id of the application for which to upload a file"),
    upload_files: List[UploadFile] = File(..., description="The application files to be uploaded"),
    content_length: int = Header(...),
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.APPLICATIONS_EDIT)),
):
    """
    Upload application files using an authenticated user token.
    """
    logger.debug(f"Preparing to receive upload files for {application_id=}")

    if content_length > settings.MAX_UPLOAD_FILE_SIZE:
        message = f"Uploaded files cannot exceed {settings.MAX_UPLOAD_FILE_SIZE} bytes."
        logger.warning(message)
        raise HTTPException(
            status_code=status.HTTP_413_REQUEST_ENTITY_TOO_LARGE,
            detail=message,
        )

    identity_claims = IdentityClaims.from_token_payload(token_payload)
    application = await _fetch_application_by_id(application_id)
    check_owner(application.application_owner_email, identity_claims.email, application_id, "application")

    ApplicationFiles.get_from_upload_files(upload_files).write_to_s3(application_id)

    update_query = (
        applications_table.update()
        .where(applications_table.c.id == application_id)
        .values(dict(application_uploaded=True))
    )

    logger.trace(f"update_query = {render_sql(update_query)}")
    await database.execute(update_query)


@router.patch(
    "/applications/{application_id}/upload/individually",
    status_code=status.HTTP_204_NO_CONTENT,
    responses={
        204: {"description": "File(s) was(were) patched successfully"},
        403: {"description": "User does not own application"},
    },
)
async def update_application_source_file(
    application_id: int = Query(..., description="id of the application for which to upload a file"),
    source_file: Union[UploadFile, None] = File(default=None),
    config_file: Union[UploadFile, None] = File(default=None),
    template_files: Union[List[UploadFile], None] = None,
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.APPLICATIONS_EDIT)),
):
    """Update the application files individually."""
    identity_claims = IdentityClaims.from_token_payload(token_payload)
    application = await _fetch_application_by_id(application_id)
    check_owner(application.application_owner_email, identity_claims.email, application_id, "application")

    # TODO: limit by file size
    ApplicationFiles(
        application_config=config_file.file.read().decode("utf-8") if config_file is not None else None,
        application_source_file=source_file.file.read().decode("utf-8") if source_file is not None else None,
        application_templates={
            PurePath(template_file.filename).name: template_file.file.read().decode("utf-8")
            for template_file in template_files
        }
        if template_files is not None
        else {},
    ).write_to_s3(application_id, remove_previous_files=False)

    return FastAPIResponse(status_code=status.HTTP_204_NO_CONTENT)


@router.delete(
    "/applications/{application_id}/upload",
    status_code=status.HTTP_204_NO_CONTENT,
    description="Endpoint for deleting application files",
)
async def applications_delete_upload(
    application_id: int = Query(..., description="id of the application for which to delete the file"),
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.APPLICATIONS_EDIT)),
):
    """
    Delete application files using an authenticated user token.
    """
    logger.debug(f"Preparing to delete files for {application_id=}")

    identity_claims = IdentityClaims.from_token_payload(token_payload)
    application = await _fetch_application_by_id(application_id)
    check_owner(application.application_owner_email, identity_claims.email, application_id, "application")

    if not application.application_uploaded:
        logger.debug(f"Trying to delete an applications that was not uploaded ({application_id=})")
        return FastAPIResponse(status_code=status.HTTP_204_NO_CONTENT)

    ApplicationFiles.delete_from_s3(application_id)

    update_query = (
        applications_table.update()
        .where(applications_table.c.id == application_id)
        .values(dict(application_uploaded=False))
    )

    logger.trace(f"update_query = {render_sql(update_query)}")
    await database.execute(update_query)

    return FastAPIResponse(status_code=status.HTTP_204_NO_CONTENT)


@router.delete(
    "/applications/{application_id}",
    status_code=status.HTTP_204_NO_CONTENT,
    description="Endpoint to delete application",
)
async def application_delete(
    application_id: int = Query(..., description="id of the application to delete"),
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.APPLICATIONS_EDIT)),
):
    """
    Delete application from the database and S3 given its id.
    """
    identity_claims = IdentityClaims.from_token_payload(token_payload)
    application = await _fetch_application_by_id(application_id)
    check_owner(application.application_owner_email, identity_claims.email, application_id, "application")

    logger.debug(f"Orphaning job_scripts rendered from application {application_id=}")
    update_query = (
        job_scripts_table.update()
        .where(job_scripts_table.c.application_id == application_id)
        .values(dict(application_id=None))
    )

    logger.trace(f"update_query = {render_sql(update_query)}")
    await database.execute(update_query)

    logger.debug(f"Preparing to delete {application_id=} from the database and S3")

    where_stmt = applications_table.c.id == application_id
    delete_query = applications_table.delete().where(where_stmt)
    logger.trace(f"delete_query = {render_sql(delete_query)}")

    await database.execute(delete_query)

    ApplicationFiles.delete_from_s3(application_id)

    return FastAPIResponse(status_code=status.HTTP_204_NO_CONTENT)


@router.delete(
    "/applications",
    status_code=status.HTTP_204_NO_CONTENT,
    description="Endpoint to delete application by identifier",
)
async def application_delete_by_identifier(
    identifier: str = Query(..., description="identifier of the application to delete"),
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.APPLICATIONS_EDIT)),
):
    """
    Delete application from the database and S3 given it's identifier.
    """
    logger.debug(f"Preparing to delete {identifier=} from the database and S3")

    where_stmt = applications_table.c.application_identifier == identifier
    get_query = applications_table.select().where(where_stmt)
    logger.trace(f"get_query = {render_sql(get_query)}")

    raw_application = await database.fetch_one(get_query)
    if not raw_application:
        message = f"Application {identifier=} not found."
        logger.warning(message)
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=message,
        )
    application = ApplicationPartialResponse.parse_obj(raw_application)

    identity_claims = IdentityClaims.from_token_payload(token_payload)
    check_owner(application.application_owner_email, identity_claims.email, application.id, "application")

    delete_query = applications_table.delete().where(where_stmt)
    logger.trace(f"delete_query = {render_sql(delete_query)}")

    await database.execute(delete_query)

    ApplicationFiles.delete_from_s3(application.id)

    return FastAPIResponse(status_code=status.HTTP_204_NO_CONTENT)


@router.get(
    "/applications",
    description="Endpoint to list applications",
    responses=ok_response(ApplicationPartialResponse),
)
async def applications_list(
    user: bool = Query(False),
    all: bool = Query(False),
    include_archived: bool = Query(False),
    search: Optional[str] = Query(None),
    sort_field: Optional[str] = Query(None),
    sort_ascending: bool = Query(True),
    pagination: Pagination = Depends(),
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.APPLICATIONS_VIEW)),
):
    """
    List all applications.
    """
    logger.debug("Preparing to list the Applications")

    identity_claims = IdentityClaims.from_token_payload(token_payload)
    query = applications_table.select()
    if user:
        query = query.where(applications_table.c.application_owner_email == identity_claims.email)
    if not all:
        query = query.where(not_(applications_table.c.application_identifier.is_(None)))
    if not include_archived:
        query = query.where(not_(applications_table.c.is_archived))
    if search is not None:
        query = query.where(search_clause(search, searchable_fields))
    if sort_field is not None:
        query = query.order_by(sort_clause(sort_field, sortable_fields, sort_ascending))

    logger.trace(f"query = {render_sql(query)}")
    return await package_response(ApplicationPartialResponse, query, pagination)


@router.get(
    "/applications/{application_identification}",
    description="Endpoint to return an application given the id",
    response_model=ApplicationResponse,
    dependencies=[Depends(guard.lockdown(Permissions.APPLICATIONS_VIEW))],
)
async def applications_get_by_id(
    application_identification: Union[int, str] = Query(None),
):
    """
    Return the application given it's id (when int) or identifier (when str).
    """
    logger.debug(f"Getting {application_identification=}")

    if isinstance(application_identification, int):
        query = applications_table.select().where(applications_table.c.id == application_identification)
    elif isinstance(application_identification, str):
        query = applications_table.select().where(
            applications_table.c.application_identifier == application_identification
        )
    else:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="application_identification must be provided as int or str",
        )

    logger.trace(f"get_query = {render_sql(query)}")

    application_data = await database.fetch_one(query)
    if not application_data:
        message = f"Application {application_identification=} not found."
        logger.warning(message)
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=message,
        )

    logger.trace(f"Application data: {dict(application_data)}")

    if application_data["application_uploaded"]:
        response = ApplicationResponse(
            **application_data,
            **ApplicationFiles.get_from_s3(application_data["id"]).dict(
                by_alias=True,
                exclude_defaults=True,
                exclude_unset=True,
            ),
        )
    else:
        response = ApplicationResponse(**application_data)

    return response


@router.put(
    "/applications/{application_id}",
    status_code=status.HTTP_200_OK,
    description="Endpoint to update an application given the id",
    response_model=ApplicationPartialResponse,
)
async def application_update(
    application_id: int,
    application: ApplicationUpdateRequest,
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.APPLICATIONS_EDIT)),
):
    """
    Update an application given it's id.
    """
    logger.debug(f"Preparing to update {application_id=}")
    identity_claims = IdentityClaims.from_token_payload(token_payload)
    old_application = await _fetch_application_by_id(application_id)
    check_owner(old_application.application_owner_email, identity_claims.email, application_id, "application")

    update_query = (
        applications_table.update()
        .where(applications_table.c.id == application_id)
        .values(application.dict(exclude_unset=True))
        .returning(applications_table)
    )
    logger.trace(f"update_query = {render_sql(update_query)}")

    async with database.transaction():
        try:
            application_data = await database.fetch_one(update_query)

        except INTEGRITY_CHECK_EXCEPTIONS as e:
            raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail=str(e))

    return application_data


def include_router(app):
    """
    Include the router for this module in the app.
    """
    app.include_router(router)



"""Router for the JobScript resource."""
from typing import Optional

from armasec import TokenPayload
from fastapi import APIRouter, Depends, File, HTTPException, Query
from fastapi import Response as FastAPIResponse
from fastapi import UploadFile, status
from fastapi.responses import PlainTextResponse
from loguru import logger
from sqlalchemy import join, not_, select
from sqlalchemy.sql import func

from jobbergate_api.apps.applications.application_files import ApplicationFiles
from jobbergate_api.apps.applications.models import applications_table
from jobbergate_api.apps.applications.schemas import ApplicationResponse
from jobbergate_api.apps.job_scripts.job_script_files import (
    JOBSCRIPTS_MAIN_FILE_FOLDER,
    JobScriptCreationError,
    JobScriptFiles,
)
from jobbergate_api.apps.job_scripts.models import job_scripts_table, searchable_fields, sortable_fields
from jobbergate_api.apps.job_scripts.schemas import (
    JobScriptCreateRequest,
    JobScriptResponse,
    JobScriptUpdateRequest,
)
from jobbergate_api.apps.job_submissions.models import job_submissions_table
from jobbergate_api.apps.permissions import Permissions, check_owner
from jobbergate_api.pagination import Pagination, ok_response, package_response
from jobbergate_api.security import IdentityClaims, guard
from jobbergate_api.storage import (
    INTEGRITY_CHECK_EXCEPTIONS,
    database,
    render_sql,
    search_clause,
    sort_clause,
)

router = APIRouter()


async def _fetch_job_script_by_id(job_script_id: int) -> JobScriptResponse:
    """
    Fetch a job_script from the database by its id.
    """
    select_query = job_scripts_table.select().where(job_scripts_table.c.id == job_script_id)
    raw_job_script = await database.fetch_one(select_query)

    if not raw_job_script:
        message = f"Job Script with {job_script_id=} was not found."
        logger.error(message)
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=message)

    return JobScriptResponse.parse_obj(raw_job_script)


@router.post(
    "/job-scripts",
    status_code=status.HTTP_201_CREATED,
    response_model=JobScriptResponse,
    description="Endpoint for job_script creation",
)
async def job_script_create(
    job_script: JobScriptCreateRequest,
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.JOB_SCRIPTS_EDIT)),
):
    """
    Create a new job script.

    Make a post request to this endpoint with the required values to create a new job script.
    """
    logger.debug(f"Creating {job_script=}")

    jobscript_files = None
    application_name = None
    if job_script.application_id is not None:
        select_query = applications_table.select().where(applications_table.c.id == job_script.application_id)
        logger.trace(f"select_query = {render_sql(select_query)}")

        raw_application = await database.fetch_one(select_query)

        if not raw_application:
            message = f"Application with id={job_script.application_id} not found."
            logger.warning(message)
            raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=message)

        application = ApplicationResponse.parse_obj(raw_application)
        application_name = application.application_name

        if application.application_uploaded is False:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail=f"Application with id={job_script.application_id} was not uploaded.",
            )
        try:
            jobscript_files = JobScriptFiles.render_from_application(
                application_files=ApplicationFiles.get_from_s3(application.id),
                user_supplied_parameters=job_script.param_dict or {},
                sbatch_params=job_script.sbatch_params or [],
            )
        except JobScriptCreationError as e:
            message = str(e)
            logger.error(message)
            raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=message)

    identity_claims = IdentityClaims.from_token_payload(token_payload)

    create_dict = dict(
        **{
            k: v
            for (k, v) in job_script.dict(exclude_unset=True).items()
            if k not in ("param_dict", "sbatch_params")
        },
        job_script_owner_email=identity_claims.email,
    )

    logger.debug("Inserting job_script")

    async with database.transaction():
        try:
            insert_query = job_scripts_table.insert().returning(job_scripts_table)
            logger.trace(f"insert_query = {render_sql(insert_query)}")
            job_script_data = await database.fetch_one(query=insert_query, values=create_dict)

            if job_script_data is None:
                message = "An error occurred when inserting the JobScript at the database."
                logger.error(message)
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail=message,
                )

            if jobscript_files:
                jobscript_files.write_to_s3(job_script_data["id"])

        except INTEGRITY_CHECK_EXCEPTIONS as e:
            logger.error(f"Reverting database transaction: {str(e)}")
            raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(e))

        except KeyError as e:
            logger.error(f"Reverting database transaction: {str(e)}")
            raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(e))

    logger.debug(f"Job-script created: {dict(job_script_data)}")

    response = JobScriptResponse(
        **job_script_data,
        application_name=application_name,
        job_script_files=jobscript_files,
    )
    return response


@router.get(
    "/job-scripts/{job_script_id}",
    description="Endpoint to get a job_script",
    response_model=JobScriptResponse,
    dependencies=[Depends(guard.lockdown(Permissions.JOB_SCRIPTS_VIEW))],
)
async def job_script_get(job_script_id: int = Query(...)):
    """
    Return the job_script given its id.
    """
    logger.debug(f"Getting {job_script_id=}")

    query = (
        select([job_scripts_table, applications_table.c.application_name])
        .select_from(
            join(
                job_scripts_table,
                applications_table,
                applications_table.columns.id == job_scripts_table.columns.application_id,
                isouter=True,
            )
        )
        .where(job_scripts_table.c.id == job_script_id)
    )
    logger.trace(f"get_query = {render_sql(query)}")
    job_script = await database.fetch_one(query)

    if not job_script:
        message = f"JobScript with id={job_script_id} not found."
        logger.warning(message)
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=message,
        )

    try:
        response = JobScriptResponse(
            **job_script,
            job_script_files=JobScriptFiles.get_from_s3(job_script_id),
        )
    except (KeyError, ValueError):
        message = f"JobScript file not found for id={job_script_id}."
        logger.warning(message)
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=message,
        )

    logger.debug(f"Job-script data: {response}")

    return response


@router.post(
    "/job-scripts/{job_script_id}/upload",
    status_code=status.HTTP_200_OK,
    description="Endpoint to upload a new job script file.",
)
async def job_script_create_file_content(
    job_script_id: int = Query(...),
    job_script_file: UploadFile = File(...),
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.JOB_SCRIPTS_EDIT)),
):
    """Create a job script file (only if the job script has none)."""
    logger.debug(f"Creating the main file for {job_script_id=}")

    identity_claims = IdentityClaims.from_token_payload(token_payload)
    job_script = await _fetch_job_script_by_id(job_script_id)
    check_owner(job_script.job_script_owner_email, identity_claims.email, job_script_id, "job_script")

    jobscript_files = JobScriptFiles.get_from_single_upload_file(job_script_file)
    jobscript_files.write_to_s3(job_script_id, remove_previous_files=True)

    update_query = (
        job_scripts_table.update()
        .where(job_scripts_table.c.id == job_script_id)
        .values(updated_at=func.now())
    )
    await database.execute(update_query)
    logger.debug(f"Success creating job script files from single upload file for {job_script_id=}")
    return dict(message=f"Successfully uploaded {job_script_file.filename} for {job_script_id=}")


@router.patch(
    "/job-scripts/{job_script_id}/upload",
    status_code=status.HTTP_204_NO_CONTENT,
    description="Endpoint to replace a job script file.",
)
async def job_script_replace_file_content(
    job_script_id: int = Query(...),
    job_script_file: UploadFile = File(...),
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.JOB_SCRIPTS_EDIT)),
):
    """Replace the content on a job script file."""
    logger.debug(f"Replacing the main file from {job_script_id=}")

    identity_claims = IdentityClaims.from_token_payload(token_payload)
    job_script = await _fetch_job_script_by_id(job_script_id)
    check_owner(job_script.job_script_owner_email, identity_claims.email, job_script_id, "job_script")

    file_manager = JobScriptFiles.file_manager_factory(job_script_id)
    file_content = job_script_file.file.read().decode("utf-8")

    for s3_path in file_manager.keys():
        root_dir = s3_path.parts[0]
        if root_dir == JOBSCRIPTS_MAIN_FILE_FOLDER:
            file_manager[s3_path] = file_content
            update_query = (
                job_scripts_table.update()
                .where(job_scripts_table.c.id == job_script_id)
                .values(updated_at=func.now())
            )
            await database.execute(update_query)
            logger.debug(f"Success replacing the main file from {job_script_id=}")
            return None

    message = f"Main file from {job_script_id=} was not found."
    logger.warning(message)
    raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=message)


@router.get(
    "/job-scripts/{job_script_id}/download",
    status_code=status.HTTP_200_OK,
    description="Endpoint to download a job script file.",
    dependencies=[Depends(guard.lockdown(Permissions.JOB_SCRIPTS_VIEW))],
)
async def job_script_download_file(job_script_id: int = Query(...)):
    """Download the job script file."""
    logger.debug(f"Downloading main file from {job_script_id=}")

    query = job_scripts_table.select().where(job_scripts_table.c.id == job_script_id)
    logger.trace(f"get_query = {render_sql(query)}")
    job_script = await database.fetch_one(query)

    if not job_script:
        message = f"JobScript with id={job_script_id} was not found."
        logger.warning(message)
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=message)

    file_manager = JobScriptFiles.file_manager_factory(job_script_id)

    for s3_path in file_manager.keys():
        root_dir = s3_path.parts[0]
        if root_dir == JOBSCRIPTS_MAIN_FILE_FOLDER:
            file_content = file_manager[s3_path]
            filename = s3_path.relative_to(JOBSCRIPTS_MAIN_FILE_FOLDER).as_posix()
            return PlainTextResponse(file_content, media_type="text/plain", headers={"filename": filename})

    message = f"Main file from {job_script_id=} was not found."
    logger.warning(message)
    raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=message)


@router.get(
    "/job-scripts",
    description="Endpoint to list job_scripts",
    responses=ok_response(JobScriptResponse),
)
async def job_script_list(
    pagination: Pagination = Depends(),
    all: Optional[bool] = Query(False),
    include_archived: bool = Query(False),
    search: Optional[str] = Query(None),
    sort_field: Optional[str] = Query(None),
    from_application_id: Optional[int] = Query(
        None,
        description="Filter job-scripts by the application-id they were created from.",
    ),
    sort_ascending: bool = Query(True),
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.JOB_SCRIPTS_VIEW)),
):
    """
    List job_scripts for the authenticated user.

    Note::

       Use responses instead of response_model to skip a second round of validation and serialization. This
       is already happening in the ``package_response`` method. So, we uses ``responses`` so that FastAPI
       can generate the correct OpenAPI spec but not post-process the response.
    """
    logger.debug("Preparing to list job-scripts")

    query = select([job_scripts_table, applications_table.c.application_name]).select_from(
        join(
            job_scripts_table,
            applications_table,
            applications_table.c.id == job_scripts_table.c.application_id,
            isouter=True,
        )
    )
    identity_claims = IdentityClaims.from_token_payload(token_payload)
    if not all:
        query = query.where(job_scripts_table.c.job_script_owner_email == identity_claims.email)
    if not include_archived:
        query = query.where(not_(job_scripts_table.c.is_archived))
    if from_application_id is not None:
        query = query.where(job_scripts_table.c.application_id == from_application_id)
    if search is not None:
        query = query.where(search_clause(search, searchable_fields))
    if sort_field is not None:
        query = query.order_by(sort_clause(sort_field, sortable_fields, sort_ascending))
    else:
        query = query.order_by(job_scripts_table.c.id.asc())

    logger.debug(f"Query = {render_sql(query)}")
    return await package_response(JobScriptResponse, query, pagination)


@router.delete(
    "/job-scripts/{job_script_id}",
    status_code=status.HTTP_204_NO_CONTENT,
    description="Endpoint to delete job script",
)
async def job_script_delete(
    job_script_id: int = Query(..., description="id of the job script to delete"),
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.JOB_SCRIPTS_EDIT)),
):
    """
    Delete job_script given its id.
    """
    identity_claims = IdentityClaims.from_token_payload(token_payload)
    job_script = await _fetch_job_script_by_id(job_script_id)
    check_owner(job_script.job_script_owner_email, identity_claims.email, job_script_id, "job_script")

    logger.debug(f"Orphaning job_submissions submitted from job_script {job_script_id=}")
    update_query = (
        job_submissions_table.update()
        .where(job_submissions_table.c.job_script_id == job_script_id)
        .values(dict(job_script_id=None))
    )
    logger.trace(f"update_query = {render_sql(update_query)}")
    await database.execute(update_query)

    logger.debug(f"Preparing to delete {job_script_id=}")
    delete_query = job_scripts_table.delete().where(job_scripts_table.c.id == job_script_id)
    logger.trace(f"delete_query = {render_sql(delete_query)}")
    await database.execute(delete_query)

    try:
        JobScriptFiles.delete_from_s3(job_script_id)
    except KeyError:
        # There is no need to raise an error if we try to delete a file that does not exist
        logger.warning(f"Tried to delete {job_script_id=}, but it was not found on S3.")

    return FastAPIResponse(status_code=status.HTTP_204_NO_CONTENT)


@router.put(
    "/job-scripts/{job_script_id}",
    status_code=status.HTTP_200_OK,
    description="Endpoint to update a job_script given the id",
    response_model=JobScriptResponse,
)
async def job_script_update(
    job_script_id: int,
    job_script: JobScriptUpdateRequest,
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.JOB_SCRIPTS_EDIT)),
):
    """
    Update a job_script given its id.
    """
    logger.debug(f"Updating {job_script_id=}")
    identity_claims = IdentityClaims.from_token_payload(token_payload)
    old_job_script = await _fetch_job_script_by_id(job_script_id)
    check_owner(old_job_script.job_script_owner_email, identity_claims.email, job_script_id, "job_script")

    update_query = (
        job_scripts_table.update()
        .where(job_scripts_table.c.id == job_script_id)
        .values(
            job_script.dict(exclude_unset=True, exclude={"job_script_files"}),
        )
        .returning(job_scripts_table)
    )
    logger.trace(f"update_query = {render_sql(update_query)}")

    async with database.transaction():
        try:
            result = await database.fetch_one(update_query)
        except INTEGRITY_CHECK_EXCEPTIONS as e:
            logger.error(f"Reverting database transaction: {str(e)}")
            raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail=str(e))

        if result is None:
            message = f"JobScript with id={job_script_id} not found."
            logger.warning(message)
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=message,
            )

        job_script_response = dict(result)

        try:
            if job_script.job_script_files:
                job_script.job_script_files.write_to_s3(job_script_id)
                job_script_response["job_script_files"] = job_script.job_script_files
            else:
                job_script_response["job_script_files"] = JobScriptFiles.get_from_s3(job_script_id)
        except (KeyError, ValueError) as e:
            logger.error(f"Reverting database transaction: {str(e)}")
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=f"File for JobScript with id={job_script_id} not found in S3.",
            )

    return job_script_response


def include_router(app):
    """
    Include the router for this module in the app.
    """
    app.include_router(router)



"""
Router for the JobSubmission resource.
"""
from typing import Any, Dict, List, Optional

from armasec import TokenPayload
from fastapi import APIRouter, Body, Depends, HTTPException, Query, status
from loguru import logger
from sqlalchemy import join, select

from jobbergate_api.apps.applications.models import applications_table
from jobbergate_api.apps.job_scripts.job_script_files import JobScriptFiles
from jobbergate_api.apps.job_scripts.models import job_scripts_table
from jobbergate_api.apps.job_submissions.constants import JobSubmissionStatus
from jobbergate_api.apps.job_submissions.models import (
    job_submissions_table,
    searchable_fields,
    sortable_fields,
)
from jobbergate_api.apps.job_submissions.properties_parser import get_job_properties_from_job_script
from jobbergate_api.apps.job_submissions.schemas import (
    ActiveJobSubmission,
    JobSubmissionCreateRequest,
    JobSubmissionResponse,
    JobSubmissionUpdateRequest,
    PendingJobSubmission,
)
from jobbergate_api.apps.permissions import Permissions, check_owner
from jobbergate_api.email_notification import notify_submission_rejected
from jobbergate_api.pagination import Pagination, ok_response, package_response
from jobbergate_api.security import IdentityClaims, guard
from jobbergate_api.storage import (
    INTEGRITY_CHECK_EXCEPTIONS,
    database,
    render_sql,
    search_clause,
    sort_clause,
)

router = APIRouter()


async def _fetch_job_submission_by_id(job_submission_id: int) -> JobSubmissionResponse:
    """
    Fetch a job_submission from the database by its id.
    """
    select_query = job_submissions_table.select().where(job_submissions_table.c.id == job_submission_id)
    raw_job_submission = await database.fetch_one(select_query)

    if not raw_job_submission:
        message = f"Job Submission with {job_submission_id=} was not found."
        logger.error(message)
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND, detail=message)

    return JobSubmissionResponse.parse_obj(raw_job_submission)


@router.post(
    "/job-submissions",
    status_code=201,
    description="Endpoint for job_submission creation",
    response_model=JobSubmissionResponse,
)
async def job_submission_create(
    job_submission: JobSubmissionCreateRequest,
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.JOB_SUBMISSIONS_EDIT)),
):
    """
    Create a new job submission.

    Make a post request to this endpoint with the required values to create a new job submission.
    """
    logger.debug(f"Creating {job_submission=}")

    identity_claims = IdentityClaims.from_token_payload(token_payload)
    client_id = job_submission.client_id or identity_claims.client_id
    if client_id is None:
        message = "Could not find a client_id in the request body or auth token."
        logger.warning(message)
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=message,
        )

    new_job_submission_data: Dict[str, Any] = dict(
        **job_submission.dict(exclude_unset=True),
        job_submission_owner_email=identity_claims.email,
        status=JobSubmissionStatus.CREATED,
    )
    if job_submission.client_id is None:
        new_job_submission_data.update(client_id=client_id)

    exec_dir = new_job_submission_data.pop("execution_directory", None)
    if exec_dir is not None:
        new_job_submission_data.update(execution_directory=str(exec_dir))

    select_query = job_scripts_table.select().where(job_scripts_table.c.id == job_submission.job_script_id)
    logger.trace(f"job_scripts select_query = {render_sql(select_query)}")
    raw_job_script = await database.fetch_one(select_query)

    if not raw_job_script:
        message = f"JobScript id={job_submission.job_script_id} not found."
        logger.warning(message)
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=message,
        )

    try:
        job_properties = get_job_properties_from_job_script(
            job_submission.job_script_id, **job_submission.execution_parameters.dict(exclude_unset=True)
        )
        new_job_submission_data["execution_parameters"] = job_properties.dict(exclude_unset=True)
    except Exception as e:
        message = f"Error extracting execution parameters from job script: {str(e)}"
        logger.error(message)
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=message)

    logger.debug("Inserting job-submission")

    async with database.transaction():
        try:
            insert_query = job_submissions_table.insert().returning(job_submissions_table)
            logger.trace(f"job_submissions insert_query = {render_sql(insert_query)}")
            job_submission_data = await database.fetch_one(query=insert_query, values=new_job_submission_data)

        except INTEGRITY_CHECK_EXCEPTIONS as e:
            logger.error(f"Reverting database transaction: {str(e)}")
            raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(e))

    logger.debug(f"Job-submission created: {job_submission_data=}")

    return JobSubmissionResponse(
        **job_submission_data, job_script_name=raw_job_script.get("job_script_name")  # type: ignore
    )


@router.get(
    "/job-submissions/{job_submission_id}",
    description="Endpoint to get a job_submission",
    response_model=JobSubmissionResponse,
    dependencies=[Depends(guard.lockdown(Permissions.JOB_SUBMISSIONS_VIEW))],
)
async def job_submission_get(job_submission_id: int = Query(...)):
    """
    Return the job_submission given it's id.
    """
    logger.debug(f"Getting {job_submission_id=}")

    query = (
        select([job_submissions_table, job_scripts_table.c.job_script_name])
        .select_from(
            join(
                job_submissions_table,
                job_scripts_table,
                job_submissions_table.c.job_script_id == job_scripts_table.c.id,
                isouter=True,
            )
        )
        .where(job_submissions_table.c.id == job_submission_id)
    )
    logger.trace(f"query = {render_sql(query)}")
    job_submission_data = await database.fetch_one(query)

    if not job_submission_data:
        message = f"JobSubmission with id={job_submission_id} not found."
        logger.warning(message)
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=message,
        )

    logger.debug(f"Job-submission data: {job_submission_data=}")

    return JobSubmissionResponse(**job_submission_data)  # type: ignore


@router.get(
    "/job-submissions",
    description="Endpoint to list job_submissions",
    responses=ok_response(JobSubmissionResponse),
)
async def job_submission_list(
    pagination: Pagination = Depends(),
    all: Optional[bool] = Query(
        None,
        description="If supplied, do not limit job_submissions to only the current user",
    ),
    slurm_job_ids: Optional[str] = Query(
        None,
        description="Comma-separated list of slurm-job-ids to match active job_submissions",
    ),
    submit_status: Optional[JobSubmissionStatus] = Query(
        None,
        description="Limit results to those with matching status",
    ),
    search: Optional[str] = Query(None),
    sort_field: Optional[str] = Query(None),
    from_job_script_id: Optional[int] = Query(
        None,
        description="Filter job-submissions by the job-script-id they were created from.",
    ),
    sort_ascending: bool = Query(True),
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.JOB_SUBMISSIONS_VIEW)),
):
    """
    List job_submissions for the authenticated user.
    """
    logger.debug("Fetching job submissions")
    identity_claims = IdentityClaims.from_token_payload(token_payload)
    logger.debug(f"Extracted identity claims from token: {identity_claims}")
    query = select([job_submissions_table, job_scripts_table.c.job_script_name]).select_from(
        join(
            job_submissions_table,
            job_scripts_table,
            job_submissions_table.c.job_script_id == job_scripts_table.c.id,
            isouter=True,
        )
    )

    logger.debug("Building query")
    if submit_status:
        query = query.where(job_submissions_table.c.status == submit_status)

    if not all:
        query = query.where(job_submissions_table.c.job_submission_owner_email == identity_claims.email)

    if from_job_script_id is not None:
        query = query.where(job_submissions_table.c.job_script_id == from_job_script_id)
    if slurm_job_ids is not None and slurm_job_ids != "":
        try:
            job_ids = [int(i) for i in slurm_job_ids.split(",")]
        except Exception:
            raise HTTPException(
                status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
                detail="Invalid slurm_job_ids param. Must be a comma-separated list of integers",
            )
        query = query.where(job_submissions_table.c.slurm_job_id.in_(job_ids))
    if search is not None:
        query = query.where(search_clause(search, searchable_fields))
    if sort_field is not None:
        query = query.order_by(sort_clause(sort_field, sortable_fields, sort_ascending))
    else:
        query = query.order_by(job_submissions_table.c.id.asc())

    logger.trace(f"Query built as: {render_sql(query)}")

    logger.debug("Awaiting query and response package")
    response = await package_response(JobSubmissionResponse, query, pagination)
    logger.debug(f"Response built as: {response}")
    return response


@router.delete(
    "/job-submissions/{job_submission_id}",
    status_code=status.HTTP_204_NO_CONTENT,
    description="Endpoint to delete job submission",
)
async def job_submission_delete(
    job_submission_id: int = Query(..., description="id of the job submission to delete"),
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.JOB_SUBMISSIONS_EDIT)),
):
    """
    Delete job_submission given its id.
    """
    logger.debug(f"Deleting {job_submission_id=}")

    identity_claims = IdentityClaims.from_token_payload(token_payload)
    job_submission = await _fetch_job_submission_by_id(job_submission_id)
    check_owner(
        job_submission.job_submission_owner_email, identity_claims.email, job_submission_id, "job_submission"
    )

    delete_query = job_submissions_table.delete().where(job_submissions_table.c.id == job_submission_id)
    logger.trace(f"delete_query = {render_sql(delete_query)}")
    await database.execute(delete_query)


@router.put(
    "/job-submissions/{job_submission_id}",
    status_code=status.HTTP_200_OK,
    description="Endpoint to update a job_submission given the id",
    response_model=JobSubmissionResponse,
)
async def job_submission_update(
    job_submission_id: int,
    job_submission: JobSubmissionUpdateRequest,
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.JOB_SUBMISSIONS_EDIT)),
):
    """
    Update a job_submission given its id.
    """
    logger.debug(f"Updating {job_submission_id=}")

    identity_claims = IdentityClaims.from_token_payload(token_payload)
    old_job_submission = await _fetch_job_submission_by_id(job_submission_id)
    check_owner(
        old_job_submission.job_submission_owner_email,
        identity_claims.email,
        job_submission_id,
        "job_submission",
    )

    update_dict = job_submission.dict(exclude_unset=True)
    exec_dir = update_dict.pop("execution_directory", None)
    if exec_dir is not None:
        update_dict.update(execution_directory=str(exec_dir))

    update_query = (
        job_submissions_table.update()
        .where(job_submissions_table.c.id == job_submission_id)
        .values(update_dict)
        .returning(job_submissions_table)
    )
    logger.trace(f"update_query = {render_sql(update_query)}")
    async with database.transaction():
        try:
            job_submission_data = await database.fetch_one(update_query)

        except INTEGRITY_CHECK_EXCEPTIONS as e:
            logger.error(f"Reverting database transaction: {str(e)}")
            raise HTTPException(status_code=status.HTTP_409_CONFLICT, detail=str(e))

        if not job_submission_data:
            message = f"JobSubmission with id={job_submission_id} not found."
            logger.warning(message)
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail=message,
            )

    return JobSubmissionResponse(**job_submission_data)  # type: ignore


# The "agent" routes are used for agents to fetch pending job submissions and update their statuses
@router.get(
    "/job-submissions/agent/pending",
    description="Endpoint to list pending job_submissions for the requesting client",
    response_model=List[PendingJobSubmission],
)
async def job_submissions_agent_pending(
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.JOB_SUBMISSIONS_VIEW)),
):
    """
    Get a list of pending job submissions for the cluster-agent.
    """
    logger.debug("Agent is requesting a list of pending job submissions")

    identity_claims = IdentityClaims.from_token_payload(token_payload)
    if identity_claims.client_id is None:
        message = "Access token does not contain a `client_id`. Cannot fetch pending submissions"
        logger.warning(message)
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=message)

    logger.info(f"Fetching newly created job_submissions for client_id: {identity_claims.client_id}")

    query = (
        select(
            columns=[
                job_submissions_table.c.id,
                job_submissions_table.c.job_submission_name,
                job_submissions_table.c.job_submission_owner_email,
                job_submissions_table.c.execution_directory,
                job_submissions_table.c.execution_parameters,
                job_submissions_table.c.job_script_id,
                job_scripts_table.c.job_script_name,
                applications_table.c.application_name,
            ]
        )
        .select_from(job_submissions_table.join(job_scripts_table).join(applications_table))
        .where(
            job_submissions_table.c.status == JobSubmissionStatus.CREATED,
        )
        .where(
            job_submissions_table.c.client_id == identity_claims.client_id,
        )
    )
    logger.trace(f"query = {render_sql(query)}")
    rows = await database.fetch_all(query)

    response = []
    missing_ids = set()

    for row in rows:
        try:
            jobscript_id = int(row["job_script_id"])
            job_script_files = JobScriptFiles.get_from_s3(jobscript_id)
        except (KeyError, ValueError) as e:
            logger.error(f"Error getting files for {jobscript_id=}: {str(e)}")
            missing_ids.add(jobscript_id)
        else:
            response.append(PendingJobSubmission(**row, job_script_files=job_script_files))

    if missing_ids:
        message = f"JobScript file(s) not found, the missing ids are: {', '.join(map(str, missing_ids))}"
        logger.error(message)
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=message,
        )

    return response


@router.put(
    "/job-submissions/agent/{job_submission_id}",
    status_code=200,
    description="Endpoint for an agent to update the status of a job_submission",
    response_model=JobSubmissionResponse,
)
async def job_submission_agent_update(
    job_submission_id: int,
    new_status: str = Body(..., embed=True),
    slurm_job_id: Optional[int] = Body(None, embed=True),
    report_message: Optional[str] = Body(None, embed=True),
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.JOB_SUBMISSIONS_EDIT)),
):
    """
    Update a job_submission with a new status.

    Make a put request to this endpoint with the new status to update a job_submission.
    """
    logger.debug(f"Agent is requesting to update {job_submission_id=}")

    identity_claims = IdentityClaims.from_token_payload(token_payload)
    if identity_claims.client_id is None:
        logger.error("Access token does not contain a client_id")
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Access token does not contain a `client_id`. Cannot update job_submission",
        )

    logger.info(
        f"Setting status to: {new_status} "
        f"for job_submission: {job_submission_id} "
        f"on client_id: {identity_claims.client_id}"
    )

    update_values: Dict[str, Any] = dict(status=new_status)
    if slurm_job_id is not None:
        update_values.update(slurm_job_id=slurm_job_id)

    if report_message:
        update_values.update(report_message=report_message)

    update_query = (
        job_submissions_table.update()
        .where(job_submissions_table.c.id == job_submission_id)
        .where(job_submissions_table.c.client_id == identity_claims.client_id)
        .values(**update_values)
        .returning(job_submissions_table)
    )
    logger.trace(f"update_query = {render_sql(update_query)}")
    job_submission_data = await database.fetch_one(update_query)

    if job_submission_data is None:
        message = (
            f"JobSubmission with id={job_submission_id} "
            f"and client_id={identity_claims.client_id} not found."
        )
        logger.warning(message)
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=(message),
        )

    if report_message and new_status == JobSubmissionStatus.REJECTED:
        notify_submission_rejected(
            job_submission_id, report_message, job_submission_data["job_submission_owner_email"]
        )

    return JobSubmissionResponse(**job_submission_data)  # type: ignore


@router.get(
    "/job-submissions/agent/active",
    description="Endpoint to list active job_submissions for the requesting client",
    response_model=List[ActiveJobSubmission],
)
async def job_submissions_agent_active(
    token_payload: TokenPayload = Depends(guard.lockdown(Permissions.JOB_SUBMISSIONS_VIEW)),
):
    """
    Get a list of active job submissions for the cluster-agent.
    """
    logger.debug("Agent is requesting a list of active job submissions")

    identity_claims = IdentityClaims.from_token_payload(token_payload)
    if identity_claims.client_id is None:
        logger.error("Access token does not contain a client_id")
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Access token does not contain a `client_id`. Cannot fetch pending submissions",
        )

    logger.info(f"Fetching active job_submissions for client_id: {identity_claims.client_id}")

    query = (
        job_submissions_table.select()
        .where(job_submissions_table.c.status == JobSubmissionStatus.SUBMITTED)
        .where(job_submissions_table.c.client_id == identity_claims.client_id)
    )
    logger.trace(f"query = {render_sql(query)}")

    rows = await database.fetch_all(query)
    return rows


def include_router(app):
    """
    Include the router for this module in the app.
    """
    app.include_router(router)



"""
Test s3 manager.
"""

import pytest
from boto3 import client

from jobbergate_api.s3_manager import engine_factory


class TestEngineFactory:
    """
    Test the engine factory.
    """

    @pytest.fixture(scope="class")
    def client(self):
        """
        Fixture with a dummy client for testing.
        """
        return client("s3")

    @pytest.fixture(scope="class")
    def s3_engine(self, client):
        """
        Fixture with a dummy s3 engine factory for testing.
        """
        return engine_factory(
            s3_client=client,
            bucket_name="test-bucket",
            prefix="test-dir",
        )

    def test_client(self, s3_engine, client):
        """
        Assert the client has the expected value.
        """
        assert s3_engine.s3_client == client

    def test_bucket_name(self, s3_engine):
        """
        Assert the bucket_name has the expected value.
        """
        assert s3_engine.bucket_name == "test-bucket"

    def test_prefix(self, s3_engine):
        """
        Assert the prefix has the expected value.
        """
        assert s3_engine.prefix == "test-dir"



"""
JobScript resource schema.
"""
from datetime import datetime
from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from jobbergate_api.apps.job_scripts.job_script_files import JobScriptFiles
from jobbergate_api.meta_mapper import MetaField, MetaMapper

job_script_meta_mapper = MetaMapper(
    id=MetaField(
        description="The unique database identifier for the instance",
        example=101,
    ),
    created_at=MetaField(
        description="The timestamp for when the instance was created",
        example="2021-12-28 23:13:00",
    ),
    updated_at=MetaField(
        description="The timestamp for when the instance was last updated",
        example="2021-12-28 23:52:00",
    ),
    job_script_name=MetaField(
        description="The unique name of the instance",
        example="test-job-script-88",
    ),
    job_script_description=MetaField(
        description="A text field providing a human-friendly description of the job_script",
        example="This job_scripts runs an Foo job using the bar variant",
    ),
    job_script_data_as_string=MetaField(
        description="The job_script itself. This is base64 encoded. Example below is decoded for clarity.",
        example=" ".join(
            [
                '{"application.sh": "#!/bin/bash\n\n#SBATCH --job-name=rats\n#SBATCH',
                "--partition=partition1\n#SBATCH --output=sample-%j.out\n\n\nsource",
                "/opt/openfoam8/etc/bashrc\n\nexport",
                'PATH=$PATH:/opt/openfoam8/platforms/linux64GccDPInt32Opt/bin\n\n\nblockMesh\nsimpleFoam"}',
            ]
        ),
    ),
    job_script_owner_email=MetaField(
        description="The email of the owner/creator of the instance",
        example="tucker@omnivector.solutions",
    ),
    application_id=MetaField(
        description="The foreign-key to the application from which this instance was created",
        example=71,
    ),
    sbatch_params=MetaField(
        description="SBATCH parameters to inject into the job_script",
        example=["alpha", "beta", "delta"],
    ),
    param_dict=MetaField(
        description="Parameters to use when rendering the job_script jinja2 template",
        example={"param1": 7, "param2": 13},
    ),
    is_archived=MetaField(
        description="Indicates if the job_script has been archived.",
        example=False,
    ),
)


class JobScriptCreateRequest(BaseModel):
    """
    Request model for creating JobScript instances.
    """

    job_script_name: str
    job_script_description: Optional[str]
    application_id: Optional[int]
    sbatch_params: Optional[List[str]]
    param_dict: Optional[Dict[str, Any]]

    class Config:
        schema_extra = job_script_meta_mapper


class JobScriptUpdateRequest(BaseModel):
    """
    Request model for updating JobScript instances.
    """

    job_script_name: Optional[str]
    job_script_description: Optional[str]
    job_script_files: Optional[JobScriptFiles]
    sbatch_params: Optional[List[str]]
    param_dict: Optional[str]
    is_archived: Optional[bool]

    class Config:
        schema_extra = job_script_meta_mapper


class JobScriptResponse(BaseModel):
    """
    Complete model to match database for the JobScript resource.

    Notice this model does not include job_script_files from S3.
    """

    id: Optional[int] = None
    created_at: Optional[datetime] = None
    updated_at: Optional[datetime] = None
    job_script_name: str
    job_script_description: Optional[str] = None
    job_script_files: Optional[JobScriptFiles] = None
    job_script_owner_email: str
    application_name: Optional[str] = None
    application_id: Optional[int] = None
    is_archived: bool

    class Config:
        orm_mode = True
        schema_extra = job_script_meta_mapper



"""
Test the properties parser module.
"""
import contextlib
import inspect
from argparse import ArgumentParser
from pathlib import Path
from typing import MutableMapping
from unittest import mock

import pytest
from bidict import bidict
from pydantic import ValidationError

from jobbergate_api.apps.job_scripts.job_script_files import JobScriptFiles
from jobbergate_api.apps.job_submissions.properties_parser import (
    _IDENTIFICATION_FLAG,
    _INLINE_COMMENT_MARK,
    ArgumentParserCustomExit,
    _clean_jobscript,
    _clean_line,
    _flagged_line,
    build_mapping_sbatch_to_slurm,
    build_parser,
    convert_sbatch_to_slurm_api,
    get_job_parameters,
    get_job_properties_from_job_script,
    jobscript_to_dict,
    sbatch_to_slurm_mapping,
    split,
)
from jobbergate_api.apps.job_submissions.schemas import JobProperties


def test_identification_flag():
    """
    Check if the value of the identification flag is the same codded in the tests.
    """
    assert _IDENTIFICATION_FLAG == "#SBATCH"


def test_inline_comment_mark():
    """
    Check if the inline comment mark is the same codded in the tests.
    """
    assert _INLINE_COMMENT_MARK == "#"


@pytest.mark.parametrize(
    "line, desired_value",
    (
        ("#SBATCH -abc", True),
        ("##SBATCH -abc", False),
        ("run python application", False),
        ("# A comment", False),
    ),
)
def test_flagged_line(line, desired_value):
    """
    Check if the flagged lines are identified properly.
    """
    actual_value = _flagged_line(line)
    assert actual_value == desired_value


@pytest.mark.parametrize(
    "line, desired_value",
    (
        ("#SBATCH", ""),
        ("#SBATCH#SBATCH", ""),
        ("#SBATCH -abc # A comment", "-abc"),
        ("#SBATCH --abc=0 # A comment", "--abc=0"),
        ("#SBATCH --abc 0 # A comment", "--abc 0"),
        ("#SBATCH -a 0 # A comment", "-a 0"),
        ("#SBATCH -a=0 # A comment", "-a=0"),
        ("#SBATCH -a = 0 # A comment", "-a = 0"),
        ("#SBATCH -a=C&Aaccount # A comment", "-a=C&Aaccount"),
    ),
)
def test_clean_line(line, desired_value):
    """
    Check if the provided lines are cleaned properly.

    I.e., identification flag, inline comments, and comment mark are all removed.
    """
    actual_value = _clean_line(line)
    assert actual_value == desired_value


@pytest.mark.parametrize(
    "line, desired_value",
    (
        ("--help", ["--help"]),
        ("--abc=0", ["--abc=0"]),
        ("-J job_name", ["-J", "job_name"]),
        ("-v -J job_name", ["-v", "-J", "job_name"]),
        ("-J job_name -v", ["-J", "job_name", "-v"]),
        ("-a=0", ["-a=0"]),
        ("-a = 0", ["-a", "=", "0"]),
        ("-a 0", ["-a", "0"]),
    ),
)
def test_split_line(line, desired_value):
    """
    Base check for shlex.split, to record the current behavior.
    """
    actual_value = split(line)
    assert actual_value == desired_value


@pytest.fixture
def dummy_slurm_script():
    """
    Provide a dummy job script for testing.
    """
    # TODO: DRY, integrate this one with conftest/dummy_template_source
    return inspect.cleandoc(
        """
        #!/bin/bash
        #SBATCH                                 # Empty line
        #SBATCH --no-kill --no-requeue          # Flagged params
        #SBATCH -n 4 -A <account>               # Multiple args per line
        #SBATCH --job-name "serial_job_test"    # Job name
        #SBATCH --mail-type=END,FAIL            # Mail events (NONE, BEGIN, END, FAIL, ALL)
        #SBATCH --mail-user=email@somewhere.com # Where to send mail
        #SBATCH --mem=1gb                       # Job memory request
        #SBATCH --time=00:05:00                 # Time limit hrs:min:sec
        #SBATCH --output=serial_test_%j.log     # Standard output and error log
        #SBATCH --wait-all-nodes 0              #
        #SBATCH --acctg-freq=task=60,energy=30,network=30,filesystem=60
        #SBATCH --comment="Hi, I am a comment"
        pwd; hostname; date

        module load python

        echo "Running plot script on a single CPU core"

        python /data/training/SLURM/plot_template.py

        date
        """
    )


def test_clean_jobscript(dummy_slurm_script):
    """
    Check if all sbatch parameters are correctly extracted from a job script.

    This operation combines many of the functions tested above (filter,
    clean, and slit the parameters on each line).
    """
    desired_list = [
        "--no-kill",
        "--no-requeue",
        "-n",
        "4",
        "-A",
        "<account>",
        "--job-name",
        "serial_job_test",
        "--mail-type=END,FAIL",
        "--mail-user=email@somewhere.com",
        "--mem=1gb",
        "--time=00:05:00",
        "--output=serial_test_%j.log",
        "--wait-all-nodes",
        "0",
        "--acctg-freq=task=60,energy=30,network=30,filesystem=60",
        "--comment=Hi, I am a comment",
    ]
    actual_list = list(_clean_jobscript(dummy_slurm_script))
    assert actual_list == desired_list


@pytest.mark.parametrize("item", sbatch_to_slurm_mapping)
class TestSbatchToSlurmList:
    """
    Test all the fields of the objects SbatchToSlurm that are stored in the list sbatch_to_slurm.
    """

    def test_slurmrestd_var_name__is_string(self, item):
        """
        Check if the field slurmrestd_var_name is a string.
        """
        assert isinstance(item.slurmrestd_var_name, str)

    def test_slurmrestd_var_name__only_contains_letters(self, item):
        """
        Check if the field slurmrestd_var_name contains only underscores and letters.
        """
        if item.slurmrestd_var_name:
            assert item.slurmrestd_var_name.replace("_", "").isalpha()

    def test_sbatch__is_string(self, item):
        """
        Check if the field sbatch is a string.
        """
        assert isinstance(item.sbatch, str)

    def test_sbatch__starts_with_double_hyphen(self, item):
        """
        Check if the field sbatch starts with a double hyphen.
        """
        assert item.sbatch.startswith("--")

    def test_sbatch__is_not_empty(self, item):
        """
        Check if the field sbatch is not empty by asserting that it has more than three characters.

        Notice it is supposed to start with a double hyphen.
        """
        assert len(item.sbatch) >= 3

    def test_sbatch__only_contains_letters(self, item):
        """
        Check if the field sbatch contains only hyphens and letters.
        """
        assert item.sbatch.replace("-", "").isalpha()

    def test_sbatch_short__is_string(self, item):
        """
        Check if the optional field sbatch_short  is a string.
        """
        assert isinstance(item.sbatch_short, str)

    def test_sbatch_short__starts_with_hyphen(self, item):
        """
        Check if the optional field sbatch_short starts with a hyphen.
        """
        if item.sbatch_short:
            assert item.sbatch_short.startswith("-")

    def test_sbatch_short__is_not_empty(self, item):
        """
        Check if of the optional field sbatch_short is equal to two, since it should be a hyphen and a letter.
        """
        if item.sbatch_short:
            assert len(item.sbatch_short) == 2

    def test_sbatch_short__only_contains_letters(self, item):
        """
        Check if the optional field sbatch_short contains only hyphens and letters.
        """
        if item.sbatch_short:
            assert item.sbatch_short.replace("-", "").isalpha()

    def test_argparser_param__is_mutable_mapping(self, item):
        """
        Check if the field argparser_param is a MutableMapping.
        """
        assert isinstance(item.argparser_param, MutableMapping)

    def test_argparser_param__is_valid_for_parser(self, item):
        """
        Check if the field argparser_param can be added in a parser.
        """
        if item.argparser_param:
            args = (i for i in (item.sbatch_short, item.sbatch) if i)
            parser = ArgumentParserCustomExit()
            parser.add_argument(*args, **item.argparser_param)


@pytest.mark.parametrize("field", ["slurmrestd_var_name", "sbatch", "sbatch_short"])
def test_sbatch_to_slurm_list__contains_only_unique_values(field):
    """
    Test that any given field has no duplicated values for all parameters stored at sbatch_to_slurm.

    This aims to avoid ambiguity at the SBATCH argparser and the two-way mapping between
    SBATCH and Slurm Rest API namespaces.
    """
    list_of_values = [getattr(i, field) for i in sbatch_to_slurm_mapping if bool(getattr(i, field))]

    assert len(list_of_values) == len(set(list_of_values))


class TestArgumentParserCustomExit:
    """
    Test the custom error handling implemented over the built-in argument parser.
    """

    @pytest.fixture(scope="module")
    def parser(self):
        """
        Support the tests in this class bt providing an instance of the parser.
        """
        parser = ArgumentParserCustomExit()
        parser.add_argument("--foo", type=int)
        parser.add_argument("--bar", action="store_true")
        return parser

    def test_argument_parser_success(self, parser):
        """
        Test the base case, where the arguments are successfully parsed and converted to the expected type.
        """
        args = parser.parse_args("--foo=10 --bar".split())
        assert {"foo": 10, "bar": True} == vars(args)

    def test_argument_parser_raise_value_error_when_value_is_missing(self, parser):
        """
        Test that ValueError is raised when the value for one parameter is missing.
        """
        with pytest.raises(ValueError, match="error: argument --foo: expected one argument"):
            parser.parse_args("--foo".split())

    def test_argument_parser_raise_value_error_in_case_of_wrong_type(self, parser):
        """
        Test that ValueError is raised when the value for some parameter is incompatible with its type.
        """
        with pytest.raises(ValueError, match="error: argument --foo: invalid int value:"):
            parser.parse_args("--foo some_text".split())


def test_build_parser():
    """
    Test if build_parser runs with no problem and returns the correct type.
    """
    parser = build_parser()
    assert isinstance(parser, ArgumentParser)


def test_build_mapping_sbatch_to_slurm():
    """
    Test if build_mapping_sbatch_to_slurm runs with no problem and returns the correct type.
    """
    mapping = build_mapping_sbatch_to_slurm()
    assert isinstance(mapping, bidict)


def test_jobscript_to_dict__success(dummy_slurm_script):
    """
    Test if the SBATCH parameters are properly extracted from a job script.

    They are expected to be returned a dictionary mapping parameters to their value.
    """
    desired_dict = {
        "no_kill": True,
        "requeue": False,
        "account": "<account>",
        "job_name": "serial_job_test",
        "mail_type": "END,FAIL",
        "mail_user": "email@somewhere.com",
        "mem": "1gb",
        "ntasks": 4,
        "output": "serial_test_%j.log",
        "time": "00:05:00",
        "wait_all_nodes": 0,
        "acctg_freq": "task=60,energy=30,network=30,filesystem=60",
        "comment": "Hi, I am a comment",
    }

    actual_dict = jobscript_to_dict(dummy_slurm_script)

    assert actual_dict == desired_dict


def test_jobscript_to_dict__raises_exception_for_unknown_parameter():
    """
    Test if jobscript_to_dict raises a ValueError when facing unknown parameters.
    """
    with pytest.raises(ValueError, match="Unrecognized SBATCH arguments:"):
        jobscript_to_dict("#SBATCH --foo\n#SBATCH --bar=0")


@pytest.mark.parametrize(
    "item",
    filter(lambda i: getattr(i, "slurmrestd_var_name", False), sbatch_to_slurm_mapping),
)
def test_convert_sbatch_to_slurm_api__success(item):
    """
    Test if the keys in a dictionary are properly renamed from SBATCH to Slurm Rest API namespace.

    Notice the values should not be affected.
    """
    desired_dict = {item.slurmrestd_var_name: None}

    sbatch_name = item.sbatch.lstrip("-").replace("-", "_")
    actual_dict = convert_sbatch_to_slurm_api({sbatch_name: None})

    assert actual_dict == desired_dict


def test_convert_sbatch_to_slurm_api__raises_exception_for_unknown_parameter():
    """
    Test the conversion of dict keys from SBATCH to Slurm Rest API raises KeyError when facing unknown names.
    """
    with pytest.raises(KeyError, match="Impossible to convert from SBATCH to Slurm REST API:"):
        convert_sbatch_to_slurm_api(dict(foo=0, bar=1))


def test_get_job_parameters(dummy_slurm_script):
    """
    Test if all SBATCH parameters are properly extracted from a given job script.

    The name of each of them is mapped to Slurm Rest API namespace and returned
    in a dictionary. Notice get_job_parameters accepts extra keywords that can be
    used to overwrite the values from the job script.
    """
    desired_dict = {
        "no_kill": True,
        "requeue": False,
        "account": "<account>",
        "name": "serial_job_test",
        "mail_type": "END,FAIL",
        "mail_user": "email@somewhere.com",
        "memory_per_node": "1gb",
        "tasks": 4,
        "standard_output": "serial_test_%j.log",
        "time_limit": "00:05:00",
        "wait_all_nodes": 0,
        "account_gather_frequency": "task=60,energy=30,network=30,filesystem=60",
        "comment": "Hi, I am a comment",
    }

    actual_dict = get_job_parameters(dummy_slurm_script)

    assert desired_dict == actual_dict


@pytest.fixture
def mock_job_script_files():
    """
    Create a dummy job script file with the given content and mock JobScriptFiles.get_from_s3.
    """

    @contextlib.contextmanager
    def _helper(main_file_content):
        main_file_name = Path("jobbergate.sh")
        job_script_files = JobScriptFiles(
            main_file_path=main_file_name,
            files={main_file_name: main_file_content},
        )
        with mock.patch(
            "jobbergate_api.apps.job_submissions.properties_parser.JobScriptFiles.get_from_s3"
        ) as mocked:
            mocked.return_value = job_script_files
            yield mocked

    return _helper


class TestGetJobPropertiesFromJobScript:
    """
    Test the get_job_properties_from_job_script function.

    It covers job properties obtained from the job script and from the users when creating a job submission.
    """

    def test_base_case(self, mock_job_script_files):
        """
        Base case, not SBATCH parameters on the job script and no extra parameters.
        """
        job_script_id = 1
        desired_job_properties = JobProperties()

        with mock_job_script_files("almost-empty-file") as mocked:
            actual_job_properties = get_job_properties_from_job_script(job_script_id)
            mocked.assert_called_once_with(job_script_id)

        assert actual_job_properties == desired_job_properties

    def test_properties_only_from_file(self, mock_job_script_files):
        """
        Job properties are obtained only from the job script.
        """
        job_script_id = 1
        desired_job_properties = JobProperties(exclusive="exclusive", name="test-test")

        with mock_job_script_files("#SBATCH --exclusive\n#SBATCH -J test-test") as mocked:
            actual_job_properties = get_job_properties_from_job_script(job_script_id)
            mocked.assert_called_once_with(job_script_id)

        assert actual_job_properties == desired_job_properties

    def test_properties_only_from_arguments(self, mock_job_script_files):
        """
        Job properties are obtained only from the extra arguments.
        """
        job_script_id = 1

        job_properties = dict(exclusive="exclusive", name="test-test")
        desired_job_properties = JobProperties.parse_obj(job_properties)

        with mock_job_script_files("almost-empty-file") as mocked:
            actual_job_properties = get_job_properties_from_job_script(
                job_script_id,
                **job_properties,
            )
            mocked.assert_called_once_with(job_script_id)

        assert actual_job_properties == desired_job_properties

    def test_properties_priority_when_both_are_provided(self, mock_job_script_files):
        """
        Job properties are obtained from the job script and from the extra arguments.

        The ones from the extra arguments have priority over the ones from the job script.
        """
        job_script_id = 1

        job_properties = dict(exclusive="exclusive", name="high-priority-test")
        desired_job_properties = JobProperties.parse_obj(job_properties)

        with mock_job_script_files("#SBATCH --exclusive\n#SBATCH -J test-test") as mocked:
            actual_job_properties = get_job_properties_from_job_script(
                job_script_id,
                **job_properties,
            )
            mocked.assert_called_once_with(job_script_id)

        assert actual_job_properties == desired_job_properties

    def test_properties_fails_with_unknown_field(self, mock_job_script_files):
        """
        The function fails when the extra arguments contain unknown fields.
        """
        job_script_id = 1

        with mock_job_script_files("almost-empty-file") as mocked:
            with pytest.raises(ValidationError, match="1 validation error for JobProperties"):
                get_job_properties_from_job_script(job_script_id, foo="bar")
            mocked.assert_called_once_with(job_script_id)


class TestBidictMapping:
    """
    Integration test with the requirement bidict (used for two-way mapping).
    """

    @pytest.fixture
    def dummy_mapping(self):
        """
        Provide a dummy dictionary used for testing.
        """
        return {f"key_{i}": f"value_{i}" for i in range(5)}

    def test_bidict__is_mutable_mapping(self):
        """
        Check if bidict implements all the necessary protocols to be a MutableMapping.
        """
        assert issubclass(bidict, MutableMapping)

    def test_bidict__can_be_compared_to_a_dictionary(self, dummy_mapping):
        """
        Check if bidict can be really comparable to a dictionary.
        """
        assert dummy_mapping == bidict(dummy_mapping)

    def test_bidict__can_be_compared_to_a_dictionary_inverse(self, dummy_mapping):
        """
        Check if bidict can be comparable to a dictionary.

        This time checking its inverse capability (i.e., swapping keys and values).
        """
        desired_value = {value: key for key, value in dummy_mapping.items()}

        assert desired_value == bidict(dummy_mapping).inverse


class TestExclusiveParameter:
    """
    --exclusive is a special SBATCH parameter that can be used in some different ways.

    See the details:
    1. --exclusive as a flag, meaning the value 'exclusive' should be recovered for slurmd.
    2. --exclusive=<value>, meaning the value <value> should be recovered for slurmd.
    According to the Slurm documentation, the value can be 'user' or 'mcs'.
    3. On top of that, 'exclusive' is expected to be set to 'oversubscribe' when the
    flag --oversubscribe is used.

    Note: When both are used, the last of them takes precedence.
    """

    def test_empty_jobscript(self):
        """
        Base case: no --exclusive parameter at all.
        """
        jobscript = ""

        desired_dict = {}

        actual_dict = jobscript_to_dict(jobscript)

        assert actual_dict == desired_dict

    def test_exclusive_as_a_flag(self):
        """
        Test the first scenario: --exclusive as a flag.
        """
        jobscript = "#SBATCH --exclusive"

        desired_dict = {"exclusive": "exclusive"}

        actual_dict = jobscript_to_dict(jobscript)

        assert actual_dict == desired_dict

    @pytest.mark.parametrize("exclusive_value", ["user", "mcs", "exclusive", "oversubscribe"])
    def test_exclusive_with_string_value(self, exclusive_value):
        """
        Test the second scenario: --exclusive=<value>.
        """
        jobscript = f"#SBATCH --exclusive={exclusive_value}"

        desired_dict = {"exclusive": exclusive_value}

        actual_dict = jobscript_to_dict(jobscript)

        assert actual_dict == desired_dict

    def test_exclusive_with_incorrect_value(self):
        """
        Test the second scenario with an incorrect value, ValueError should be raised.
        """
        jobscript = "#SBATCH --exclusive=test-test"

        with pytest.raises(ValueError, match="invalid choice: 'test-test'"):
            jobscript_to_dict(jobscript)

    def test_oversubscribe(self):
        """
        Test the third scenario: --oversubscribe as a flag.
        """
        jobscript = "#SBATCH --oversubscribe"

        desired_dict = {"exclusive": "oversubscribe"}

        actual_dict = jobscript_to_dict(jobscript)

        assert actual_dict == desired_dict

    def test_oversubscribe_with_incorrect_value(self):
        """
        Test the second scenario with an incorrect value, ValueError should be raised.
        """
        jobscript = "#SBATCH --oversubscribe=test-test"

        with pytest.raises(
            ValueError, match="argument -s/--oversubscribe: ignored explicit argument 'test-test'"
        ):
            jobscript_to_dict(jobscript)

    def test_both_exclusive_and_oversubscribe_1(self):
        """
        Test that when both are used, the last of them takes precedence.

        In this case, --exclusive is the last one.
        """
        jobscript = "#SBATCH --oversubscribe\n#SBATCH --exclusive"

        desired_dict = {"exclusive": "exclusive"}

        actual_dict = jobscript_to_dict(jobscript)

        assert actual_dict == desired_dict

    def test_both_exclusive_and_oversubscribe_2(self):
        """
        Test that when both are used, the last of them takes precedence.

        In this case, --oversubscribe is the last one.
        """
        jobscript = "#SBATCH --exclusive\n#SBATCH --oversubscribe"

        desired_dict = {"exclusive": "oversubscribe"}

        actual_dict = jobscript_to_dict(jobscript)

        assert actual_dict == desired_dict


class TestRequeueParameter:
    """
    Requeue is a slurm parameter whose value comes from two sbatch flags: --requeue and --no-requeue.
    """

    def test_empty_jobscript(self):
        """
        Base case: no parameters at all.
        """
        jobscript = ""

        desired_dict = {}

        actual_dict = jobscript_to_dict(jobscript)

        assert actual_dict == desired_dict

    def test_requeue_flag(self):
        """
        Test the first scenario: --requeue.
        """
        jobscript = "#SBATCH --requeue"

        desired_dict = {"requeue": True}

        actual_dict = jobscript_to_dict(jobscript)

        assert actual_dict == desired_dict

    def test_no_requeue_flag(self):
        """
        Test the first scenario: --no-requeue.
        """
        jobscript = "#SBATCH --no-requeue"

        desired_dict = {"requeue": False}

        actual_dict = jobscript_to_dict(jobscript)

        assert actual_dict == desired_dict

    def test_both_exclusive_and_oversubscribe_1(self):
        """
        Test that when both are used, the last of them takes precedence.

        In this case, --no-requeue is the last one.
        """
        jobscript = "#SBATCH --requeue\n#SBATCH --no-requeue"

        desired_dict = {"requeue": False}

        actual_dict = jobscript_to_dict(jobscript)

        assert actual_dict == desired_dict

    def test_both_exclusive_and_oversubscribe_2(self):
        """
        Test that when both are used, the last of them takes precedence.

        In this case, --requeue is the last one.
        """
        jobscript = "#SBATCH --no-requeue\n#SBATCH --requeue"

        desired_dict = {"requeue": True}

        actual_dict = jobscript_to_dict(jobscript)

        assert actual_dict == desired_dict


@pytest.mark.parametrize(
    "input, expected_result",
    (
        (None, None),
        ("USR1@7200", "USR1@7200"),
        ("B:USR1@7200", "USR1@7200"),
        ("R:USR1@7200", "R:USR1@7200"),
        ("RB:USR1@7200", "R:USR1@7200"),
        ("BR:USR1@7200", "R:USR1@7200"),
        ("B:BBB", "BBB"),  # a silly case, but let's assert that just the B on the left side is removed
    ),
)
def test_backward_compatibility_on_signal_parameter(input, expected_result):
    """
    Test backward compatibility on signal parameter.

    It is expected to remove the 'B' from the prefix while preserving any other, like the 'R'.
    """
    actual_result = JobProperties(signal=input).signal
    assert actual_result == expected_result



"""
Provide a convenience class for managing calls to S3.
"""
from functools import lru_cache
from pathlib import Path
from typing import Tuple, Type, Union

from boto3 import client
from botocore.client import BaseClient
from file_storehouse import (
    EngineS3,
    FileManager,
    FileManagerReadOnly,
    TransformationABC,
    TransformationCodecs,
)

from jobbergate_api.config import settings

IO_TRANSFORMATIONS = (TransformationCodecs("utf-8"),)
"""
Transformations to be performed when writing/reading S3 objects.

With this, all files are encoded/decoded with utf-8.

This constant can be shared between file managers.
"""


def engine_factory(*, s3_client: BaseClient, bucket_name: str, prefix: str) -> EngineS3:
    """
    Build an engine to manipulate objects from an s3 bucket.

    :param BaseClient s3_client: S3 client from boto3
    :param str bucket_name: The name of the s3 bucket
    :param str prefix: Prefix for object search at s3 (Work directory)
    :return EngineS3: And engine to manipulate files from s3
    """
    return EngineS3(s3_client=s3_client, bucket_name=bucket_name, prefix=prefix)


@lru_cache(maxsize=128)
def file_manager_factory(
    id: int,
    *,
    s3_client: BaseClient,
    bucket_name: str,
    work_directory: Path,
    manager_cls: Union[Type[FileManagerReadOnly], Type[FileManager]],
    transformations: Tuple[TransformationABC, ...],
) -> Union[FileManagerReadOnly, FileManager]:
    """
    Build a file managers on demand.

    :param int id: identification number
    :param BaseClient s3_client: S3 client from boto3
    :param str bucket_name: The name of the s3 bucket
    :param Path work_directory: Work directory (referred as prefix at boto3 documentation)
    :param Union[FileManager, FileManagerReadOnly] manager_cls: Manager class (i/o access or just read only)
    :param Tuple[TransformationABC] transformations: I/o transformations
    :return Union[FileManager, FileManagerReadOnly]: Manager object
    """
    return manager_cls(
        engine=engine_factory(
            s3_client=s3_client,
            bucket_name=bucket_name,
            prefix=f"{work_directory}/{id}/",
        ),
        io_transformations=transformations,
    )


s3_client = client("s3", endpoint_url=settings.S3_ENDPOINT_URL)



# type: ignore
"""
Integration tests for the job properties.

This is a draft of the tests that may be implemented in the future,
it was added to the version control as a reference for some of the
job properties that were already tested here.

In order to run these tests, you need to:
* Spin up jobbergate-composed (and have cloned of cluster-agent for that)
* Add hypothesis as a dependency for Jobbergate API
* Remove the prefix `draft` from the filename, so pytest can find it

Test based on cluster-agent >= 2.1.0.
"""

import json
from datetime import datetime, timedelta

import pydantic
import pytest
import requests
from hypothesis import example, given, settings
from hypothesis import strategies as st
from jose import jwt
from loguru import logger
from starlette import status

from jobbergate_api.apps.job_submissions.schemas import JobProperties


@pytest.fixture(scope="module")
def token():
    """
    Create a JWT token for the local-user.
    """
    now = datetime.now()
    payload = {
        "exp": int(datetime.timestamp(now + timedelta(seconds=900))),
        "iat": int(datetime.timestamp(now)),
        "sun": "local-user",
    }
    return jwt.encode(payload, "supersecret", algorithm="HS256")


class SlurmJobSubmission(pydantic.BaseModel):
    """
    Specialized model for describing a request to submit a job to Slurm REST API.
    """

    script: str
    job: JobProperties


def test_integration_ping(token):
    """
    Test that slurmrestd is up and running.
    """
    response = requests.get(
        url="http://localhost:6820/slurm/v0.0.36/ping",
        headers={
            "x-slurm-user-name": "local-user",
            "x-slurm-user-token": token,
        },
    )
    response.raise_for_status()


def submit_job_to_slurm(payload_json, token):
    """
    Test helper, used to submit a job to slurmrestd.
    """
    logger.debug(f"Payload: {payload_json}")

    response = requests.post(
        url="http://localhost:6820/slurm/v0.0.36/job/submit",
        headers={
            "x-slurm-user-name": "local-user",
            "x-slurm-user-token": token,
        },
        json=payload_json,
    )
    logger.debug(f"Slurmrestd response: {response.json()}")

    return response


@settings(deadline=None, report_multiple_bugs=True)
@given(
    instance=st.builds(
        JobProperties,
        get_user_environment=st.integers(min_value=0, max_value=1),
        hold=st.booleans(),
        kill_on_invalid_dependency=st.booleans(),
        memory_per_node=st.integers(min_value=1, max_value=100),
        requeue=st.booleans(),
        spread_job=st.booleans(),
        wait_all_nodes=st.integers(min_value=0, max_value=1),  # not bool as slurm's documentation suggests
    )
)
@example(instance=JobProperties(nodes="1-2"))
@example(instance=JobProperties(nodes="2"))
@example(instance=JobProperties(memory_per_node="1MB"))
@example(instance=JobProperties(time_limit="1:00:00"))
@example(instance=JobProperties(standard_output="/tmp"))
@example(instance=JobProperties(open_mode="append"))
@example(instance=JobProperties(open_mode="truncate"))
def test_integration_job_submit__success(instance, token):
    """
    Test that we can submit a job to slurmrestd.
    """
    dummy_job_script = "#!/bin/bash"

    payload = SlurmJobSubmission(script=dummy_job_script, job=instance)
    payload_json = json.loads(payload.json(exclude_none=True))

    response = submit_job_to_slurm(payload_json, token)
    response.raise_for_status()


@pytest.mark.parametrize(
    "job_properties",
    [
        {"nodes": [1, 2, 3]},
        {"nodes": [1]},
        # {"nodes": "1-2"}, this one works
        # {"nodes": 1}, this one works
        # {"nodes": "1"}, this one works
        {"no_kill": True},
        {"no_kill": False},
        {"no_kill": 1},
        {"no_kill": 0},
        # {"no_kill": None}, this one works
        {"time_minimum": 10},
        {"time_minimum": "1:00:00"},
        {"wait_all_nodes": True},
        {"wait_all_nodes": False},
    ],
)
def test_integration_job_submit__known_errors(job_properties, token):
    """
    Test that we get a 400 error when submitting a job with invalid properties.
    """
    job_properties["get_user_environment"] = 0

    payload_json = dict(script="#!/bin/bash", job=job_properties)

    response = submit_job_to_slurm(payload_json, token)
    with pytest.raises(requests.exceptions.HTTPError):
        response.raise_for_status()
        assert response.status_code == status.HTTP_400_BAD_REQUEST



"""
Pagination feature for all endpoints.
"""

from typing import Dict, Generic, List, Optional, Type, TypeVar, Union

from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
from pydantic import BaseModel, Field
from pydantic.generics import GenericModel
from sqlalchemy import func, select

from jobbergate_api.storage import database


class Pagination(BaseModel):
    """
    Basic pagination class.
    """

    start: Optional[int] = Field(
        None,
        ge=0,
        description="""
            The page offset for items, where the first page index is 0.
            The index for the first item on each page is computed as

                index == start * limit.

            Value must be greater than or equal to 0.
        """,
    )
    limit: int = Field(
        10,
        gt=0,
        description="Limit of items to include per page. Must be greater than 0",
    )

    def dict(self):
        """
        Override default ``dict()`` behavior so that if ``page`` is None, all fields are omitted.
        """
        return dict() if self.start is None else super().dict()


class ResponsePagination(BaseModel):
    """
    A metadata model that describes pagination info.
    """

    total: int
    start: Optional[int]
    limit: Optional[int]


TResponseModel = TypeVar("TResponseModel", bound=BaseModel)


class Response(GenericModel, Generic[TResponseModel]):
    """
    An envelope for responses including the metadata for pagination.

    This is a generic response envelope that can be used for any of list of pydantic models.
    """

    results: List[TResponseModel]
    pagination: ResponsePagination


def ok_response(
    _: Type[TResponseModel],
) -> Dict[Union[int, str], Dict[str, Type[Response[TResponseModel]]]]:
    """
    Package up a response type compatible with FastAPI's ``responses`` argument.

    The typing is intense here, but is just basically saying that given some generic model input,
    return a dictionary that contains a ``Response`` for that model type.

    """
    return {200: {"model": Response[TResponseModel]}}


async def package_response(model: Type[TResponseModel], query, pagination: Pagination) -> JSONResponse:
    """
    Package the response in an envelope that includes the response and the metadata.

    Return a JSONResponse containing the packaged data.

    Structure of response is::

        {
            "results": [result0, result1, ...resultN],
            "metadata": {
                "total": N,
                "start": i,
                "limit", n,
            },
        }

    If pagination.start is None, no pagination is performed
    """
    # The explicit order_by removes any sorting added to the query for the total computation
    # total = await database.fetch_val(query.with_only_columns([func.count()]).order_by(None))
    count_query = select(func.count()).select_from(query.subquery())
    total = await database.fetch_val(query=count_query)
    if pagination.start is not None:
        query = query.limit(pagination.limit).offset(pagination.start * pagination.limit)
    raw_response = await database.fetch_all(query)
    validated_response = Response[TResponseModel](
        results=[model.parse_obj(x) for x in raw_response],
        pagination=ResponsePagination(total=total, **pagination.dict()),
    )
    jsonable_response = jsonable_encoder(validated_response)
    return JSONResponse(content=jsonable_response)



"""
Validation methods for the uploaded files.
"""

from ast import parse as ast_parse
from collections import Counter
from functools import wraps
from pathlib import PurePath
from textwrap import dedent
from traceback import format_tb
from typing import Callable, Dict, List, Union

from buzz import Buzz, DoExceptParams, require_condition
from fastapi import HTTPException, UploadFile, status
from jinja2 import Environment
from loguru import logger
from yaml import safe_load as yaml_safe_load

from jobbergate_api.apps.applications.constants import (
    APPLICATION_CONFIG_FILE_NAME,
    APPLICATION_SOURCE_FILE_NAME,
)
from jobbergate_api.apps.applications.schemas import ApplicationConfig


class UploadedFilesValidationError(Buzz):
    """Raise exception when facing any validation error on the uploaded files."""


def log_error_and_raise_http_error(params: DoExceptParams):
    """
    Work with ``UploadedFilesValidationError.handle_errors`` when something goes wrong.

    In case any error occurs in the context manager, the error is logged and
    a HTTPException is raised.

    :param DoExceptParams params: Params from handle_errors.
    :raises HTTPException: Code 422, unprocessable entity.
    """
    message_template = dedent(
        """
        {final_message}
        Error:
        ______
        {err}
        Traceback:
        ----------
        {trace}
        """
    )

    message = message_template.format(
        final_message=params.final_message,
        err=str(params.err),
        trace="\n".join(format_tb(params.trace)),
    )

    logger.error(message)

    raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=params.final_message)


def get_suffix(file: UploadFile) -> str:
    """
    Get the suffix (file extension) of an UploadFile.

    :param UploadFile file: Target file.
    :return str: File extension.
    """
    return PurePath(file.filename).suffix


CheckUploadedFilesEquation = Callable[[List[UploadFile]], None]
"""
Type alias describing the function signature used to check the uploaded files.

The functions are expected to raise UploadedFilesValidationError if any error is detected.
"""

check_uploaded_files_dispatch: List[CheckUploadedFilesEquation] = []
"""List the function used to check the uploaded files."""


def perform_all_checks_on_uploaded_files(file_list: List[UploadFile]):
    """
    Loop though the check-list for uploaded files specifications.

    Individual checks will raise UploadedFilesValidationError if the business
    rules are not met.

    :param List[UploadFile] file_list: Upload file list.
    """
    logger.debug("Preparing to perform all checks on the uploaded files")
    with UploadedFilesValidationError.handle_errors(
        "The uploaded files are invalid",
        do_except=log_error_and_raise_http_error,
        re_raise=False,
    ):
        for check in check_uploaded_files_dispatch:
            check(file_list)


def register_check_uploaded_files(checker: CheckUploadedFilesEquation) -> CheckUploadedFilesEquation:
    """
    Use this decorator to register functions to check the uploaded files.

    It creates a new entry on ``check_uploaded_files_dispatch``.

    :return CheckUploadedFilesEquation: The decorated function.
    """
    check_uploaded_files_dispatch.append(checker)

    @wraps(checker)
    def decorator(file_list: List[UploadFile]):
        return checker(file_list)

    return decorator


@register_check_uploaded_files
def check_uploaded_files_restrictions(upload_files: List[UploadFile]):
    """
    Check the list of uploaded files for filename restrictions.

    Jobbergate expects one source file, one (optional) config file and one
    or more template files.

    Raise UploadedFilesValidationError if the business rules are not met.

    :param List[UploadFile] file_list: Upload file list.
    """
    extension_counter = Counter(get_suffix(f) for f in upload_files)
    filenames = {upload.filename for upload in upload_files}

    logger.debug(f"Checking uploaded files extensions: {extension_counter=}")
    with UploadedFilesValidationError.check_expressions("Invalid file extensions") as check:
        check(
            APPLICATION_SOURCE_FILE_NAME in filenames,
            f"The application source file is missing ({APPLICATION_SOURCE_FILE_NAME})",
        )
        check(
            extension_counter[".j2"] + extension_counter[".jinja2"] >= 1,
            "missing one or more template files (.j2 and/or .jinja2)",
        )


@register_check_uploaded_files
def check_uploaded_files_content_type(file_list: List[UploadFile]):
    """
    Check the list of uploaded files to confirm they are specified as plain text.

    Raise UploadedFilesValidationError if the business rules are not met.

    :param List[UploadFile] file_list: Upload file list.
    """
    logger.debug("Checking uploaded files content types")
    expected = "text/plain"
    extension_list = ", ".join(syntax_validation_dispatch.keys())
    UploadedFilesValidationError.require_condition(
        all(f.content_type == expected for f in file_list if get_suffix(f) in syntax_validation_dispatch),
        f"The content of the files {extension_list} is expected to be {expected}",
    )


@register_check_uploaded_files
def check_uploaded_files_syntax(file_list: List[UploadFile]):
    """
    Check the list of uploaded files to verify their syntax.

    Raise UploadedFilesValidationError if the business rules are not met.

    :param List[UploadFile] file_list: Upload file list.
    """
    logger.debug("Checking uploaded files, validating source code")

    with UploadedFilesValidationError.check_expressions("Invalid syntax on uploaded file(s)") as check:
        for f in file_list:
            suffix = get_suffix(f)
            if suffix not in syntax_validation_dispatch:
                continue
            check(syntax_validation_dispatch[suffix](f.file.read()), f.filename)
            f.file.seek(0)


@register_check_uploaded_files
def check_uploaded_files_yaml_is_parsable(file_list: List[UploadFile]):
    """
    Check the list of uploaded files to verify if the yaml file agrees with the expected schema.

    Raise UploadedFilesValidationError if the business rules are not met.

    :param List[UploadFile] file_list: Upload file list.
    """
    logger.debug("Checking uploaded files, parsing yaml file")
    with UploadedFilesValidationError.handle_errors(
        "Not possible to get the configuration from the yaml file. "
        "Please, verify is all the required fields were provided."
    ):
        for f in file_list:
            if f.filename == APPLICATION_CONFIG_FILE_NAME:
                ApplicationConfig.get_from_yaml_file(f.file.read())
                f.file.seek(0)


SyntaxValidationEquation = Callable[[Union[str, bytes]], bool]
"""Type alias describing the function signature used to validate file syntax."""

syntax_validation_dispatch: Dict[str, SyntaxValidationEquation] = {}
"""Dictionary mapping file extensions to the function used to validate their syntax."""


def register_syntax_validator(*file_extensions: str):
    """
    Use this decorator to register file syntax validation functions.

    It creates a new entry on ``validation_dispatch``, mapping the equation to
    the file extensions that are provided as arguments.

    Raise ValueError if the provided file extensions do not start with a dot.

    :return ValidationEquation: The decorated function.
    """

    def decorator(validator):
        for extension in file_extensions:
            require_condition(
                extension.startswith("."),
                f"File extensions should start with a dot, got the value: {extension}",
                raise_exc_class=ValueError,
            )
            syntax_validation_dispatch[extension] = validator

        @wraps(validator)
        def wrapper(source_code):
            return validator(source_code)

        return wrapper

    return decorator


@register_syntax_validator(".py")
def is_valid_python_file(source_code: Union[str, bytes]) -> bool:
    """
    Check if a given Python source code is valid by parsing it into an AST node.

    :param Union[str, bytes] source_code: Python source code.
    :return bool: Boolean indicating if the source code is valid or not.
    """
    try:
        ast_parse(source_code)
    except Exception:
        return False
    return True


@register_syntax_validator(".yaml")
def is_valid_yaml_file(yaml_file: Union[str, bytes]) -> bool:
    """
    Check if a given YAML file is valid by parsing it with yaml.safe_load.

    :param Union[str, bytes] yaml_file: YAML file.
    :return bool: Boolean indicating if the file is valid or not.
    """
    try:
        yaml_safe_load(yaml_file)
    except Exception:
        return False
    return True


@register_syntax_validator(".j2", ".jinja2")
def is_valid_jinja2_template(template: Union[str, bytes]) -> bool:
    """
    Check if a given jinja2 template is valid by creating a Template object and trying to render it.

    :param str template: Jinja2 template.
    :return bool: Boolean indicating if the template is valid or not.
    """
    if isinstance(template, bytes):
        _template = template.decode("utf-8")
    else:
        _template = template

    try:
        Environment().parse(_template)
    except Exception:
        return False
    return True



"""
Test the module config.
"""

from itertools import combinations

import pytest

from jobbergate_api.config import Settings, check_none_or_all_keys_exist


class TestCheckNoneOrAllKeysExist:
    """
    A class containing all tests for check_none_or_all_keys_exist.
    """

    keys = {"foo", "bar", "baz", "qux"}
    missing_keys = {k + "_missing" for k in keys}

    @pytest.fixture(scope="class")
    def dummy_dict(self) -> dict:
        """
        Build a dummy dict to be used as reference in the tests.
        """
        return {k: None for k in self.keys}

    @pytest.mark.parametrize("target_keys", combinations(keys, 3))
    def test_all_keys_exist(self, dummy_dict, target_keys):
        """
        Test if it returns true when all of the target keys exist in the input dict.
        """
        assert check_none_or_all_keys_exist(dummy_dict, set(target_keys)) is True

    @pytest.mark.parametrize("target_keys", combinations(missing_keys, 3))
    def test_none_keys_exist(self, dummy_dict, target_keys):
        """
        Test if it returns true when none of the target keys exists in the input dict.
        """
        assert check_none_or_all_keys_exist(dummy_dict, set(target_keys)) is True

    @pytest.mark.parametrize("target_keys", zip(keys, missing_keys))
    def test_some_keys_exist(self, dummy_dict, target_keys):
        """
        Test if it returns false when some of the target keys are missing in the input dict.
        """
        assert check_none_or_all_keys_exist(dummy_dict, set(target_keys)) is False


class TestSendgridSettings:
    """
    A class containing all tests for the sendgrid settings.

    The goal is to test if all params are defined and the notifications can be sent
    to the users, or if none of the params are defined and notifications can not
    be sent. If just some of them are defined, RuntimeError is raised.
    """

    SENDGRID_PARAMS = {"SENDGRID_FROM_EMAIL", "SENDGRID_API_KEY"}

    def test_all_keys_exist(self):
        """
        Test scenario where all of the sendgrid params are defined.
        """
        params = {k: "foo" for k in self.SENDGRID_PARAMS}
        settings = Settings(**params)
        assert all(getattr(settings, k) == "foo" for k in self.SENDGRID_PARAMS)

    def test_none_key_exist(self):
        """
        Test scenario where none of the sendgrid params are defined.

        Note: Empty strings can be used here because Settings filters them out.
        """
        params = {k: "" for k in self.SENDGRID_PARAMS}
        settings = Settings(**params)
        assert all(getattr(settings, k) is None for k in self.SENDGRID_PARAMS)

    @pytest.mark.parametrize("missing_param", SENDGRID_PARAMS)
    def test_some_key_exist(self, missing_param):
        """
        Test scenario where some of the parameters are defined, resulting in RuntimeError.

        Note: Empty strings can be used here because Settings filters them out.
        """
        params = {k: "" if k == missing_param else "foo" for k in self.SENDGRID_PARAMS}
        with pytest.raises(
            RuntimeError,
            match="Either none or all SendGrind parameters are expected",
        ):
            Settings(**params)



"""
Describe constants for the applications module.
"""

APPLICATIONS_WORK_DIR = "applications"
APPLICATION_CONFIG_FILE_NAME = "jobbergate.yaml"
APPLICATION_SOURCE_FILE_NAME = "jobbergate.py"
APPLICATION_TEMPLATE_FOLDER = "templates"
